# Copyright 2010 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Language:
#
# Example definitions:
#
#  bit  foo[8];  /* Eight single bit values                 */
#  u8   foo;     /* Unsigned 8-bit value                    */
#  u16  foo;     /* Unsigned 16-bit value                   */
#  ul16 foo;     /* Unsigned 16-bit value (LE)              */
#  u24  foo;     /* Unsigned 24-bit value                   */
#  ul24 foo;     /* Unsigned 24-bit value (LE)              */
#  u32  foo;     /* Unsigned 32-bit value                   */
#  ul32 foo;     /* Unsigned 32-bit value (LE)              */
#  i8   foo;     /* Signed 8-bit value                      */
#  i16  foo;     /* Signed 16-bit value                     */
#  il16 foo;     /* Signed 16-bit value (LE)                */
#  i24  foo;     /* Signed 24-bit value                     */
#  il24 foo;     /* Signed 24-bit value (LE)                */
#  i32  foo;     /* Signed 32-bit value                     */
#  il32 foo;     /* Signed 32-bit value (LE)                */
#  char foo;     /* Character (single-byte                  */
#  lbcd foo;     /* BCD-encoded byte (LE)                   */
#  bbcd foo;     /* BCD-encoded byte (BE)                   */
#  char foo[8];  /* 8-char array                            */
#  struct {
#   u8 foo;
#   u16 bar;
#  } baz;        /* Structure with u8 and u16               */
#
# Example directives:
#
# #seekto 0x1AB; /* Set the data offset to 0x1AB            */
# #seek 4;       /* Set the data offset += 4                */
# #printoffset "foobar" /* Echo the live data offset,
#                          prefixed by string while parsing */
#
# Usage:
#
# Create a data definition in a string, and pass it and the data
# to parse to the parse() function.  The result is a structure with
# dict-like objects for structures, indexed by name, and lists of
# objects for arrays.  The actual data elements can be interpreted
# as integers directly (for int types).  Strings and BCD arrays
# behave as expected.

import struct
import os
import logging
import re
import warnings

import six
from builtins import bytes

from chirp import bitwise_grammar

LOG = logging.getLogger(__name__)


class ParseError(Exception):
    """Indicates an error parsing a definition"""
    pass


def byte_to_int(b):
    if six.PY3 or isinstance(b, int):
        return b
    else:
        return ord(b)


def int_to_byte(i):
    if six.PY3:
        return bytes([i])
    else:
        return chr(i)


def string_straight_encode(string):
    """str -> bytes"""
    # So, there are a lot of py2-thinking chirp drivers that
    # will do something like this:
    #
    #   memobj.name = 'foo\xff'
    #
    # because they need the string to be stored in radio memory
    # with a 0xFF byte terminating the string. If they do that in
    # py3 we will get the unicode equivalent, which is a non-ASCII
    # character. Conventional unicode wisdom would tell us we need
    # to encode() the string to get UTF-8 bytes, but that's not
    # what we want here (the radio doesn't support UTF-8 and we need
    # specific binary values in memory). Ideally we would have
    # written all of chirp with bytes() for these values, but alas.
    # We can get the intended string here by doing bytes([ord(char)]).
    return bytes(b''.join(int_to_byte(ord(b)) for b in string))


def string_straight_decode(string):
    """bytes -> str"""
    # Normally, we would want to decode bytes() to str() for py3.
    # However...chirp drivers are currently using strings with
    # hex escapes for setting binary byte values in memories.
    # Technically, this is char, which is a little more like bytes()
    # in py3, but until every driver has converted its strings to
    # bytes(), this is massively simpler. Since set_value() is
    # doing the inverse, we can do this here. If something sets
    # a char to '\xFF' it will arrive below as the unicode character,
    # and be converted to the integer/bytes value. When it is read out
    # here, we will return that same unicode character and the driver
    # will detect '\xFF' properly.
    # FIXMEPY3: Remove this and the hack below when drivers convert to
    # bytestrings.
    return ''.join(chr(byte_to_int(b)) for b in string)


def format_binary(nbits, value, pad=8):
    s = ""
    for i in range(0, nbits):
        s = "%i%s" % (value & 0x01, s)
        value >>= 1
    return "%s%s" % ((pad - len(s)) * ".", s)


def bits_between(start, end):
    bits = (1 << (int(end) - int(start))) - 1
    return bits << int(start)


def pp(structure, level=0):
    for i in structure:
        if isinstance(i, list):
            pp(i, level+2)
        elif isinstance(i, tuple):
            if isinstance(i[1], str):
                print("%s%s: %s" % (" " * level, i[0], i[1]))
            else:
                print("%s%s:" % (" " * level, i[0]))
                pp(i, level+2)
        elif isinstance(i, str):
            print("%s%s" % (" " * level, i))


def array_copy(dst, src):
    """Copy an array src into DataElement array dst"""
    if len(dst) != len(src):
        raise Exception("Arrays differ in size")

    for i in range(0, len(dst)):
        dst[i].set_value(src[i])


def bcd_to_int(bcd_array):
    """Convert an array of bcdDataElement like \x12\x34
    into an int like 1234"""
    value = 0
    for bcd in bcd_array:
        a, b = bcd.get_value()
        value = (value * 100) + (a * 10) + b
    return value


def int_to_bcd(bcd_array, value):
    """Convert an int like 1234 into bcdDataElements like "\x12\x34" """
    for i in reversed(range(0, len(bcd_array))):
        bcd_array[i].set_value(value % 100)
        value /= 100


def get_string(char_array):
    """Convert an array of charDataElements into a string"""
    return "".join([x.get_value() for x in char_array])


def set_string(char_array, string):
    """Set an array of charDataElements from a string"""
    array_copy(char_array, list(string))


class DataElement:
    _size = 1

    def __init__(self, data, offset, count=1):
        self._data = data
        self._offset = int(offset)
        self._count = count

    def _compat_bytes(self, bs, asbytes):
        if asbytes:
            return bytes(bs)
        else:
            warnings.warn('Driver is using non-byte-native get_raw()',
                          DeprecationWarning, stacklevel=3)
            return string_straight_decode(bs)

    def size(self):
        return int(self._size * 8)

    def get_offset(self):
        return int(self._offset)

    def _get_value(self, data):
        raise Exception("Not implemented")

    def get_value(self):
        value = self._data[self._offset:self._offset + self._size]
        return self._get_value(value)

    def set_value(self, value):
        raise Exception("Not implemented for %s" % self.__class__)

    def get_raw(self, asbytes=True):
        raw = self._data[self._offset:self._offset+self._size]
        return self._compat_bytes(raw, asbytes)

    def set_raw(self, data):
        if isinstance(data, str):
            warnings.warn('Driver is using non-byte-native set_raw()',
                          DeprecationWarning, stacklevel=2)
            data = string_straight_encode(data)
        self._data[self._offset] = data[:self._size]

    def get_path(self, path):
        """Retrieve a child element starting from here by symbolic path.

        Examples:
            .mystruct.foo[2].field1
            [3]
            .bar
        """
        tok = re.search('[.[]', path)
        if tok:
            tok = tok.group(0)
        if path.startswith('.'):
            return self.get_path(path[1:])
        elif path.startswith('['):
            index, rest = path.split(']', 1)
            index = int(index[1:])
            return self[index].get_path(rest)
        elif tok == '.':
            field, rest = path.split('.', 1)
            return getattr(self, field).get_path(rest)
        elif tok == '[':
            field, rest = path.split('[', 1)
            return getattr(self, field).get_path('[' + rest)
        elif path:
            return getattr(self, path)
        else:
            return self

    def fill_raw(self, byteval):
        """Fill object with bytes

        :param byteval: A bytes of length 1 to fill the object's memory with
        """
        assert isinstance(byteval, bytes) and len(byteval) == 1
        self.set_raw(byteval * (self.size() // 8))

    def __getattr__(self, name):
        raise AttributeError("Unknown attribute %s in %s" % (name,
                                                             self.__class__))

    def __repr__(self):
        return "(%s:%i bytes @ %04x)" % (self.__class__.__name__,
                                         self._size,
                                         self._offset)


class arrayDataElement(DataElement):
    def __repr__(self):
        if isinstance(self.__items[0], bcdDataElement):
            return "%i:[(%i)]" % (len(self.__items), int(self))

        if isinstance(self.__items[0], charDataElement):
            return "%i:[(%s)]" % (len(self.__items), repr(str(self))[1:-1])

        s = "%i:[" % len(self.__items)
        s += ",".join([repr(item) for item in self.__items])
        s += "]"
        return s

    def __init__(self, offset):
        self.__items = []
        self._offset = offset

    def append(self, item):
        self.__items.append(item)

    def get_value(self):
        return list(self.__items)

    def get_raw(self, asbytes=True):
        raw = [item.get_raw(asbytes=True) for item in self.__items]
        return self._compat_bytes(bytes(b''.join(raw)), asbytes)

    def set_raw(self, data):
        item_size = self.__items[0].size() // 8
        assert len(data) / item_size == len(self), 'Invalid raw data length'
        for i in range(len(self)):
            self[i].set_raw(data[i * item_size:(i + 1) * item_size])

    def __setitem__(self, index, val):
        self.__items[index].set_value(val)

    def __getitem__(self, index):
        if isinstance(index, slice):
            return self.__items[int(index.start):int(index.stop)]
        else:
            return self.__items[int(index)]

    def __len__(self):
        return len(self.__items)

    def __str__(self):
        if isinstance(self.__items[0], charDataElement):
            # NOTE: All the values should be strings coming out of py3.
            # and on py2 they could be a mix of unicode and str, the latter
            # for non-ASCII values. On py2 we can just coerce all of these
            # types to a string for compatibility.
            return "".join([str(x.get_value()) for x in self.__items])
        else:
            return str(self.__items)

    def __int__(self):
        if isinstance(self.__items[0], bcdDataElement):
            val = 0
            if isinstance(self.__items[0], bbcdDataElement):
                items = self.__items
            else:
                items = reversed(self.__items)
            for i in items:
                tens, ones = i.get_value()
                val = (val * 100) + (tens * 10) + ones
            return val
        else:
            raise ValueError("Cannot coerce this to int")

    def __set_value_bbcd(self, value):
        for i in reversed(self.__items):
            twodigits = value % 100
            value /= 100
            i.set_value(twodigits)

    def __set_value_lbcd(self, value):
        for i in self.__items:
            twodigits = value % 100
            value /= 100
            i.set_value(twodigits)

    def __set_value_char(self, value):
        if len(value) != len(self.__items):
            raise ValueError("String expects exactly %i characters, not %i" % (
                             len(self.__items), len(value)))
        for i in range(0, len(self.__items)):
            self.__items[i].set_value(value[i])

    def set_value(self, value):
        if isinstance(self.__items[0], bbcdDataElement):
            self.__set_value_bbcd(int(value))
        elif isinstance(self.__items[0], lbcdDataElement):
            self.__set_value_lbcd(int(value))
        elif isinstance(self.__items[0], charDataElement):
            self.__set_value_char(value)
        elif len(value) != len(self.__items):
            raise ValueError("Array cardinality mismatch")
        else:
            for i in range(0, len(value)):
                self.__items[i].set_value(value[i])

    def index(self, value):
        index = 0
        for i in self.__items:
            if i.get_value() == value:
                return index
            index += 1
        raise IndexError()

    def __iter__(self):
        return iter(self.__items)

    def items(self):
        index = 0
        for item in self.__items:
            yield (str(index), item)
            index += 1

    def size(self):
        size = 0
        for i in self.__items:
            size += i.size()
        return int(size)


class intDataElement(DataElement):
    def __repr__(self):
        fmt = "0x%%0%iX" % (self._size * 2)
        return fmt % int(self)

    def __int__(self):
        return self.get_value()

    def __nonzero__(self):
        return int(self) != 0

    def __invert__(self):
        return ~self.get_value()

    def __trunc__(self):
        return self.get_value()

    def __abs__(self):
        return abs(self.get_value())

    def __mod__(self, val):
        return self.get_value() % val

    def __mul__(self, val):
        return self.get_value() * val

    def __truediv__(self, val):
        return self.get_value() / val

    def __div__(self, val):
        return self.get_value() / val

    def __floordiv__(self, val):
        return self.get_value() // val

    def __add__(self, val):
        return self.get_value() + val

    def __sub__(self, val):
        return self.get_value() - val

    def __or__(self, val):
        return self.get_value() | val

    def __xor__(self, val):
        return self.get_value() ^ val

    def __and__(self, val):
        return self.get_value() & val

    def __radd__(self, val):
        return val + self.get_value()

    def __rsub__(self, val):
        return val - self.get_value()

    def __rmul__(self, val):
        return val * self.get_value()

    def __rdiv__(self, val):
        return val / self.get_value()

    def __rand__(self, val):
        return val & self.get_value()

    def __ror__(self, val):
        return val | self.get_value()

    def __rxor__(self, val):
        return val ^ self.get_value()

    def __rmod__(self, val):
        return val % self.get_value()

    def __lshift__(self, val):
        return self.get_value() << val

    def __rshift__(self, val):
        return self.get_value() >> val

    def __iadd__(self, val):
        self.set_value(self.get_value() + val)
        return self

    def __isub__(self, val):
        self.set_value(self.get_value() - val)
        return self

    def __imul__(self, val):
        self.set_value(self.get_value() * val)
        return self

    def __idiv__(self, val):
        self.set_value(self.get_value() / val)
        return self

    def __imod__(self, val):
        self.set_value(self.get_value() % val)
        return self

    def __iand__(self, val):
        self.set_value(self.get_value() & val)
        return self

    def __ior__(self, val):
        self.set_value(self.get_value() | val)
        return self

    def __ixor__(self, val):
        self.set_value(self.get_value() ^ val)
        return self

    def __index__(self):
        return abs(self)

    def __eq__(self, val):
        return self.get_value() == val

    def __ne__(self, val):
        return self.get_value() != val

    def __lt__(self, val):
        return self.get_value() < val

    def __le__(self, val):
        return self.get_value() <= val

    def __gt__(self, val):
        return self.get_value() > val

    def __ge__(self, val):
        return self.get_value() >= val

    def __bool__(self):
        return self.get_value() != 0


class u8DataElement(intDataElement):
    _size = 1

    def _get_value(self, data):
        return ord(data)

    def set_value(self, value):
        self._data[self._offset] = (int(value) & 0xFF)


class u16DataElement(intDataElement):
    _size = 2
    _endianess = ">"

    def _get_value(self, data):
        return struct.unpack(self._endianess + "H", data)[0]

    def set_value(self, value):
        self._data[self._offset] = struct.pack(self._endianess + "H",
                                               int(value) & 0xFFFF)


class ul16DataElement(u16DataElement):
    _endianess = "<"


class u24DataElement(intDataElement):
    _size = 3
    _endianess = ">"

    def _get_value(self, data):
        pre = self._endianess == ">" and b"\x00" or b""
        post = self._endianess == "<" and b"\x00" or b""
        return struct.unpack(self._endianess + "I", pre+data+post)[0]

    def set_value(self, value):
        if self._endianess == "<":
            start = 0
            end = 3
        else:
            start = 1
            end = 4
        packed = struct.pack(self._endianess + "I", int(value) & 0xFFFFFFFF)
        self._data[self._offset] = packed[start:end]


class ul24DataElement(u24DataElement):
    _endianess = "<"


class u32DataElement(intDataElement):
    _size = 4
    _endianess = ">"

    def _get_value(self, data):
        return struct.unpack(self._endianess + "I", data)[0]

    def set_value(self, value):
        self._data[self._offset] = struct.pack(self._endianess + "I",
                                               int(value) & 0xFFFFFFFF)


class ul32DataElement(u32DataElement):
    _endianess = "<"


class i8DataElement(u8DataElement):
    _size = 1

    def _get_value(self, data):
        return struct.unpack("b", data)[0]

    def set_value(self, value):
        self._data[self._offset] = struct.pack("b", int(value))


class i16DataElement(intDataElement):
    _size = 2
    _endianess = ">"

    def _get_value(self, data):
        return struct.unpack(self._endianess + "h", data)[0]

    def set_value(self, value):
        self._data[self._offset] = struct.pack(self._endianess + "h",
                                               int(value))


class il16DataElement(i16DataElement):
    _endianess = "<"


class i24DataElement(intDataElement):
    _size = 3
    _endianess = ">"

    def _get_value(self, data):
        pre = self._endianess == ">" and "\x00" or ""
        post = self._endianess == "<" and "\x00" or ""
        return struct.unpack(self._endianess + "i", pre+data+post)[0]

    def set_value(self, value):
        if self._endianess == "<":
            start = 0
            end = 3
        else:
            start = 1
            end = 4
        self._data[self._offset] = struct.pack(self._endianess + "i",
                                               int(value))[start:end]


class il24DataElement(i24DataElement):
    _endianess = "<"


class i32DataElement(intDataElement):
    _size = 4
    _endianess = ">"

    def _get_value(self, data):
        return struct.unpack(self._endianess + "i", data)[0]

    def set_value(self, value):
        self._data[self._offset] = struct.pack(self._endianess + "i",
                                               int(value))


class il32DataElement(i32DataElement):
    _endianess = "<"


class charDataElement(DataElement):
    _size = 1

    def __str__(self):
        return str(self.get_value())

    def __int__(self):
        return ord(self.get_value())

    def _get_value(self, data):
        return string_straight_decode(data)

    def set_value(self, value):
        if isinstance(value, int):
            # This is the case if bytes() are passed in to arrayDataElement
            # to set a string
            self._data[self._offset] = value
        else:
            self._data[self._offset] = string_straight_encode(value)


class bcdDataElement(DataElement):
    def __int__(self):
        tens, ones = self.get_value()
        return (tens * 10) + ones

    def set_bits(self, mask):
        self._data[self._offset] = ord(self._data[self._offset]) | int(mask)

    def clr_bits(self, mask):
        self._data[self._offset] = ord(self._data[self._offset]) & ~int(mask)

    def get_bits(self, mask):
        return ord(self._data[self._offset]) & int(mask)

    def set_raw(self, data):
        if isinstance(data, int):
            self._data[self._offset] = data & 0xFF
        elif isinstance(data, str):
            warnings.warn('Driver is using non-byte-native set_raw()',
                          DeprecationWarning, stacklevel=2)
            self._data[self._offset] = string_straight_encode(data[0])
        elif isinstance(data, bytes) and len(data) == 1:
            self._data[self._offset] = int(data[0]) & 0xFF
        else:
            raise TypeError("Unable to set bcdDataElement from type %s" %
                            type(data))

    def set_value(self, value):
        self._data[self._offset] = int("%02i" % value, 16)

    def _get_value(self, data):
        a = (ord(data) & 0xF0) >> 4
        b = ord(data) & 0x0F
        return (a, b)


class lbcdDataElement(bcdDataElement):
    _size = 1


class bbcdDataElement(bcdDataElement):
    _size = 1


class bitDataElement(intDataElement):
    _nbits = 0
    _shift = 0
    _subgen = u8DataElement  # Default to a byte

    def __repr__(self):
        fmt = "0x%%0%iX (%%sb)" % (self._size * 2)
        return fmt % (int(self), format_binary(self._nbits, self.get_value()))

    def get_value(self):
        data = self._subgen(self._data, self._offset).get_value()
        mask = bits_between(self._shift-self._nbits, self._shift)
        val = (data & mask) >> int(self._shift - self._nbits)
        return val

    def set_value(self, value):
        mask = bits_between(self._shift-self._nbits, self._shift)

        data = self._subgen(self._data, self._offset).get_value()
        data &= ~mask

        value = ((int(value) << int(self._shift-self._nbits)) & mask) | data

        self._subgen(self._data, self._offset).set_value(value)

    def size(self):
        return int(self._nbits)


class structDataElement(DataElement):
    def __repr__(self):
        s = "struct {" + os.linesep
        for prop in self._keys:
            s += "  %15s: %s%s" % (prop, repr(self._generators[prop]),
                                   os.linesep)
        s += "} %s (%i bytes at 0x%04X)%s" % (self._name,
                                              self.size() // 8,
                                              self._offset,
                                              os.linesep)
        return s

    def __init__(self, *args, **kwargs):
        self._generators = {}
        self._keys = []
        self._count = 1
        if "name" in kwargs.keys():
            self._name = kwargs["name"]
            del kwargs["name"]
        else:
            self._name = "(anonymous)"
        DataElement.__init__(self, *args, **kwargs)
        self.__init = True

    def _value(self, data, generators):
        result = {}
        for name, gen in generators.items():
            result[name] = gen.get_value(data)
        return result

    def get_value(self):
        result = []
        for i in range(0, self._count):
            result.append(self._value(self._data, self._generators[i]))

        if self._count == 1:
            return result[0]
        else:
            return result

    def __contains__(self, key):
        return key in self._generators.keys()

    def __getitem__(self, key):
        return self._generators[key]

    def __setitem__(self, key, value):
        if key in self._generators:
            self._generators[key].set_value(value)
        else:
            self._generators[key] = value
            self._keys.append(key)

    def __getattr__(self, name):
        try:
            return self._generators[name]
        except KeyError:
            LOG.error('Request for struct element %s not in %s' % (
                name, self._generators.keys()))
            raise AttributeError("No attribute %s in struct %s" % (
                name, self._name))

    def __setattr__(self, name, value):
        if "_structDataElement__init" not in self.__dict__:
            self.__dict__[name] = value
        else:
            self.__dict__["_generators"][name].set_value(value)

    def size(self):
        size = 0
        for name, gen in self._generators.items():
            if not isinstance(gen, list):
                gen = [gen]

            i = 0
            for el in gen:
                i += 1
                size += el.size()
        return int(size)

    def get_raw(self, asbytes=True):
        size = self.size() // 8
        raw = self._data[self._offset:self._offset+size]
        return self._compat_bytes(raw, asbytes)

    def set_raw(self, buffer):
        if len(buffer) != (self.size() // 8):
            raise ValueError("Struct size mismatch during set_raw()")
        if isinstance(buffer, str):
            warnings.warn('Driver is using non-byte-native set_raw()',
                          DeprecationWarning, stacklevel=2)
            buffer = string_straight_encode(buffer)
        self._data[self._offset] = buffer

    def __iter__(self):
        for item in self._generators.values():
            yield item

    def items(self):
        for key in self._keys:
            yield key, self._generators[key]


class Processor:
    _types = {
        "u8":    u8DataElement,
        "u16":   u16DataElement,
        "ul16":  ul16DataElement,
        "u24":   u24DataElement,
        "ul24":  ul24DataElement,
        "u32":   u32DataElement,
        "ul32":  ul32DataElement,
        "i8":    i8DataElement,
        "i16":   i16DataElement,
        "il16":  il16DataElement,
        "i24":   i24DataElement,
        "il24":  il24DataElement,
        "i32":   i32DataElement,
        "char":  charDataElement,
        "lbcd":  lbcdDataElement,
        "bbcd":  bbcdDataElement,
        }

    def __init__(self, data, offset, input):
        if hasattr(data, 'get_byte_compatible'):
            # bitwise uses the byte-compatible interface of MemoryMap,
            # if that is what was passed in
            data = data.get_byte_compatible()
        self._data = data
        self._offset = offset
        self._obj = None
        self._user_types = {}
        self._input = input.split('\n')

    def do_symbol(self, symdef, gen):
        name = symdef[1]
        self._generators[name] = gen

    def get_source_line(self, lineno):
        return self._input[lineno]

    def get_line_from_sym(self, sym):
        name = sym.__name__
        if ':' in name.line:
            _, num = name.line.split(':')
        else:
            num = 0
        return int(num)

    def get_line_sym(self, sym):
        num = self.get_line_from_sym(sym)
        return '%i: %s' % (num, self.get_source_line(num))

    def do_bitfield(self, dtype, bitfield):
        bytes = self._types[dtype](self._data, 0).size() // 8
        bitsleft = bytes * 8

        for _bitdef, defn in bitfield:
            name = defn[0][1]
            if name in self._generators:
                newname = '%s_%06x' % (name, self._offset)
                prevline = self._lines.get(name, 'unknown')
                LOG.error('Duplicate definition for %s on line %s; '
                          'renaming to %s (previous definition line %s)',
                          name, self.get_line_sym(defn[0]), newname, prevline)
                name = newname
            bits = int(defn[1][1])
            if bitsleft < 0:
                raise ParseError("Invalid bitfield spec")

            class bitDE(bitDataElement):
                _nbits = bits
                _shift = bitsleft
                _subgen = self._types[dtype]

            self._generators[name] = bitDE(self._data, self._offset)
            self._lines[name] = self.get_line_from_sym(defn[0])
            bitsleft -= bits

        if bitsleft:
            LOG.warn("WARNING: %i trailing bits unaccounted for in %s" %
                     (bitsleft, name))

        return bytes

    def do_bitarray(self, i, count):
        if count % 8 != 0:
            raise ValueError("bit array must be divisible by 8.")

        class bitDE(bitDataElement):
            _nbits = 1
            _shift = 8 - i % 8

        return bitDE(self._data, self._offset)

    def parse_defn(self, defn):
        dtype = defn[0]

        if defn[1][0] == "bitfield":
            size = self.do_bitfield(dtype, defn[1][1])
            count = 1
            self._offset += size
        else:
            if defn[1][0] == "array":
                sym = defn[1][1][0]
                count = int(defn[1][1][1][1])
            else:
                count = 1
                sym = defn[1]

            name = sym[1]
            if name in self._generators:
                newname = '%s_%06x' % (name, self._offset)
                prevline = self._lines.get(name, 'unknown')
                LOG.error('Duplicate definition for %s on line %s; '
                          'renaming to %s (previous definition line %s)',
                          name, self.get_line_sym(sym), newname, prevline)
                name = newname
            res = arrayDataElement(self._offset)
            size = 0
            for i in range(0, count):
                if dtype == "bit":
                    gen = self.do_bitarray(i, count)
                    self._offset += int((i+1) % 8 == 0)
                else:
                    gen = self._types[dtype](self._data, self._offset)
                    self._offset += (gen.size() // 8)
                res.append(gen)

            if count == 1:
                self._generators[name] = res[0]
            else:
                self._generators[name] = res
            self._lines[name] = self.get_line_from_sym(sym)

    def parse_struct_decl(self, struct):
        block = struct[:-1]
        if block[0][0] == "symbol":
            # This is a pre-defined struct
            block = self._user_types[block[0][1]]
        deftype = struct[-1]
        if deftype[0] == "array":
            name = deftype[1][0][1]
            count = int(deftype[1][1][1])
        elif deftype[0] == "symbol":
            name = deftype[1]
            count = 1

        result = arrayDataElement(self._offset)
        for i in range(0, count):
            element = structDataElement(self._data, self._offset, count,
                                        name=name)
            result.append(element)
            tmp = self._generators
            tmp_lines = self._lines
            self._generators = element
            self._lines = {}
            self.parse_block(block)
            self._generators = tmp
            self._lines = tmp_lines

        if count == 1:
            self._generators[name] = result[0]
        else:
            self._generators[name] = result

    def parse_struct_defn(self, struct):
        name = struct[0][1]
        block = struct[1:]
        self._user_types[name] = block

    def parse_struct(self, struct):
        if struct[0][0] == "struct_defn":
            return self.parse_struct_defn(struct[0][1])
        elif struct[0][0] == "struct_decl":
            return self.parse_struct_decl(struct[0][1])
        else:
            raise Exception("Internal error: What is `%s'?" % struct[0][0])

    def assert_negative_seek(self, message):
        warnings.warn(message, DeprecationWarning, stacklevel=6)

    def assert_unnecessary_seek(self, message):
        warnings.warn(message, DeprecationWarning, stacklevel=6)

    def parse_directive(self, directive):
        name = directive[0][0]
        value = directive[0][1][0][1]
        if name == "seekto":
            target = int(value, 0)
            if self._offset == target:
                self.assert_unnecessary_seek('Unnecessary #seekto %s' % value)
            elif target < self._offset:
                self.assert_negative_seek(
                    'Invalid negative seek from 0x%04x to 0x%04x' % (
                        self._offset, target))
            self._offset = target
        elif name == "seek":
            self._offset += int(value, 0)
        elif name == "printoffset":
            LOG.debug("%s: %i (0x%08X)" %
                      (value[1:-1], self._offset, self._offset))

    def parse_block(self, lang):
        for t, d in lang:
            if t == "struct":
                self.parse_struct(d)
            elif t == "definition":
                self.parse_defn(d)
            elif t == "directive":
                self.parse_directive(d)

    def parse(self, lang):
        self._lines = {}
        self._generators = structDataElement(self._data, self._offset)
        self.parse_block(lang)
        return self._generators


def parse(spec, data, offset=0):
    ast = bitwise_grammar.parse(spec)
    p = Processor(data, offset, spec)
    return p.parse(ast)



# Copyright 2019 Pavel Milanes, CO7WT <pavelmc@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import time

from chirp import chirp_common, directory, memmap, errors, bitwise
from chirp.settings import RadioSettingGroup, RadioSetting, \
    RadioSettingValueBoolean, RadioSettingValueList, \
    RadioSettings

LOG = logging.getLogger(__name__)

# SAMPLE MEM DUMP as sent from the radios

# FTL-1011
# 0x000000  52 f0 16 90 04 08 38 c0  00 00 00 01 00 00 00 ff  |R.....8.........|
# 0x000010  20 f1 00 20 00 00 00 20  04 47 25 04 47 25 00 00  | .. ... .G%.G%..|

# FTL-2011
# 0x000000: 50 90 21 40 04 80 fc 40  00 00 00 01 00 00 00 ff  |P.!@...@........|
# 0x000010: 20 f1 00 0b 00 00 00 0b  14 51 70 14 45 70 00 00  |.........Qp.Ep..|


MEM_FORMAT = """
#seekto 0x000;
u8 rid;                 // Radio Identification

struct {
u8 scan_time:4,         // Scan timer per channel: 0-15 (5-80msec in 5msec steps)
   unknownA:4;
bbcd if[2];             // Radio internal IF, depending on model (16.90, 21.40, 45.10, 47.90)
u8 chcount;             // how many channels are programmed
u8 scan_resume:1,           // Scan sesume: 0 = 0.5 seconds, 1 = Carrier
   priority_during_scan:1,  // Priority during scan: 0 = enabled, 1 = disabled
   priority_speed:1,        // Priority speed: 0 = slow, 1 = fast
   monitor:1,               // Monitor: 0 = enabled, 1 = disabled
   off_hook:1,              // Off hook: 0 = enabled, 1 = disabled
   home_channel:1,          // Home Channel: 0 = Scan Start ch, 1 = Priority 1ch
   talk_back:1,             // Talk Back: 0 = enabled, 1 = disabled
   tx_carrier_delay:1;      // TX carrier delay: 1 = enabled, 0 = disabled
u8 tot:4,                   // Time out timer: 16 values (0.0-7.5 in 0.5s step)
   tot_resume:2,            // Time out timer resume: 3, 2, 1, 0 => 0s, 6s, 20s, 60s
   unknownB:2;
u8 a_key:2,                 // A key function: resume: 0-3: Talkaround, High/Low, Call, Accessory
   unknownB:6;
u8 pch1;                    // Priority channel 1
u8 pch2;                    // Priority channel 1
} settings;

#seekto 0x010;
struct {
  u8 notx:1,            // 0 = Tx possible,  1 = Tx disabled
     empty:1,           // 0 = channel enabled, 1 = channed empty
     tot:1,             // 0 = tot disabled, 1 = tot enabled
     power:1,           // 0 = high, 1 = low
     bclo_cw:1,         // 0 = disabled, 1 = Busy Channel Lock out by carrier
     bclo_tone:1,       // 0 = disabled, 1 = Busy Channel Lock out by tone (set rx tone)
     skip:1,            // 0 = scan enabled, 1 = skip on scanning
     unknownA0:1;
  u8 chname;
  u8 rx_tone[2];      // empty value is \x00\x0B / disabled is \x00\x00
  u8 unknown4;
  u8 unknown5;
  u8 tx_tone[2];      // empty value is \x00\x0B / disabled is \x00\x00
  bbcd rx_freq[3];      // RX freq
  bbcd tx_freq[3];      // TX freq
  u8 unknownA[2];
} memory[24];

#seekto 0x0190;
char filename[11];

#seekto 0x19C;
u8 checksum;
"""

MEM_SIZE = 0x019C
POWER_LEVELS = [chirp_common.PowerLevel("High", watts=50),
                chirp_common.PowerLevel("Low", watts=5)]
DTCS_CODES = chirp_common.DTCS_CODES
SKIP_VALUES = ["", "S"]
LIST_BCL = ["OFF", "Carrier", "Tone"]
LIST_SCAN_RESUME = ["0.5 seconds", "Carrier drop"]
LIST_SCAN_TIME = ["%s ms" % x for x in range(5, 85, 5)]
LIST_SCAN_P_SPEED = ["Slow", "Fast"]
LIST_HOME_CHANNEL = ["Scan Start ch", "Priority 1ch"]
LIST_TOT = ["Off"] + ["%.1f s" % (x/10.0) for x in range(5, 80, 5)]
# 3, 2, 1, 0 => 0 s, 6 s, 20 s, 60 s
LIST_TOT_RESUME = ["60 s", "20 s", "6 s", "0 s"]
LIST_A_KEY = ["Talkaround", "High/Low", "Call", "Accessory"]
LIST_PCH = []  # dynamic, as depends on channel list.
# make a copy of the tones, is not funny to work with this directly
TONES = list(chirp_common.TONES)
# this old radios has not the full tone ranges in CST
invalid_tones = (
    69.3,
    159.8,
    165.5,
    171.3,
    177.3,
    183.5,
    189.9,
    196.6,
    199.5,
    206.5,
    229.1,
    245.1)

# remove invalid tones
for tone in invalid_tones:
    try:
        TONES.remove(tone)
    except:
        pass


def _set_serial(radio):
    """Set the serial protocol settings"""
    radio.pipe.timeout = 10
    radio.pipe.parity = "N"
    radio.pipe.baudrate = 9600


def _checksum(data):
    """the radio block checksum algorithm"""
    cs = 0
    for byte in data:
        cs += ord(byte)

    return cs % 256


def _update_cs(radio):
    """Update the checksum on the mmap"""
    payload = str(radio.get_mmap())[:-1]
    cs = _checksum(payload)
    radio._mmap[MEM_SIZE - 1] = cs


def _do_download(radio):
    """ The download function """
    # Get the whole 413 bytes (0x019D) bytes one at a time with plenty of time
    # to get to the user's pace

    # set serial discipline
    _set_serial(radio)

    # UI progress
    status = chirp_common.Status()
    status.cur = 0
    status.max = MEM_SIZE
    status.msg = " Press A to clone. "
    radio.status_fn(status)

    data = ""
    for i in range(0, MEM_SIZE):
        a = radio.pipe.read(1)
        if len(a) == 0:
            # error, no received data
            if len(data) != 0:
                # received some data, not the complete stream
                msg = "Just %02i bytes of the %02i received, try again." % \
                    (len(data), MEM_SIZE)
            else:
                # timeout, please retry
                msg = "No data received, try again."

            raise errors.RadioError(msg)

        data += a
        # UI Update
        status.cur = len(data)
        radio.status_fn(status)

    if len(data) != MEM_SIZE:
        msg = "Incomplete data, we need %02i but got %02i bytes." % \
            (MEM_SIZE, len(data))
        raise errors.RadioError(msg)

    if ord(data[-1]) != _checksum(data[:-1]):
        msg = "Bad checksum, please try again."
        raise errors.RadioError(msg)

    return data


def _do_upload(radio):
    """The upload function"""
    # set serial discipline
    _set_serial(radio)

    # UI progress
    status = chirp_common.Status()

    # 10 seconds timeout
    status.cur = 0
    status.max = 100
    status.msg = " Quick, press MON on the radio to start. "
    radio.status_fn(status)

    for byte in range(0, 100):
        status.cur = byte
        radio.status_fn(status)
        time.sleep(0.1)

    # real upload if user don't cancel the timeout
    status.cur = 0
    status.max = MEM_SIZE
    status.msg = " Cloning to radio... "
    radio.status_fn(status)

    # send data
    data = str(radio.get_mmap())

    # this radio has a trick, the EEPROM is an ancient SPI one, so it needs
    # some time to write, so we send every byte and then allow
    # a 0.01 seg to complete the write from the MCU to the SPI EEPROM
    c = 0
    for byte in data:
        radio.pipe.write(byte)
        time.sleep(0.01)

        # UI Update
        status.cur = c
        radio.status_fn(status)

        # counter
        c = c + 1


def _model_match(cls, data):
    """Use a experimental guess to determine if the radio you just
    downloaded or the img you opened is for this model"""

    # It's hard to tell when this radio is really this radio.
    # I use the first byte, that appears to be the ID and the IF settings

    radiod = [data[0], data[2:4]]
    return cls.finger == radiod


def bcd_to_int(data):
    """Convert an array of bcdDataElement like \x12
    into an int like 12"""
    value = 0
    a = (data & 0xF0) >> 4
    b = data & 0x0F
    value = (a * 10) + b
    return value


def int_to_bcd(data):
    """Convert a int like 94 to 0x94"""
    data, lsb = divmod(data, 10)
    data, msb = divmod(data, 10)
    res = (msb << 4) + lsb
    return res


class ftlx011(chirp_common.CloneModeRadio, chirp_common.ExperimentalRadio):
    """Vertex FTL1011/2011/7011 4/8/12/24 channels"""
    VENDOR = "Vertex Standard"
    _memsize = MEM_SIZE
    _upper = 0
    _range = []
    finger = []  # two elements rid & IF

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = \
            ('This is a experimental driver, use it on your own risk.\n'
             '\n'
             'This driver is just for the 4/12/24 channels variants of '
             'these radios, 99 channel variants are not supported yet.\n'
             '\n'
             'The 99 channel versions appears to use another mem layout.\n'
             )
        rp.pre_download = _(
            "Please follow this steps carefully:\n"
            "1 - Turn on your radio\n"
            "2 - Connect the interface cable to your radio.\n"
            "3 - Click the button on this window to start download\n"
            "    (Radio will beep and led will flash)\n"
            "4 - Then press the \"A\" button in your radio to start"
            " cloning.\n"
            "    (At the end radio will beep)\n")
        rp.pre_upload = _(
            "Please follow this steps carefully:\n"
            "1 - Turn on your radio\n"
            "2 - Connect the interface cable to your radio\n"
            "3 - Click the button on this window to start download\n"
            "    (you may see another dialog, click ok)\n"
            "4 - Radio will beep and led will flash\n"
            "5 - You will get a 10 seconds timeout to press \"MON\" before\n"
            "    data upload start\n"
            "6 - If all goes right radio will beep at end.\n"
            "After cloning remove the cable and power cycle your radio to\n"
            "get into normal mode.\n")
        return rp

    def get_features(self):
        """Return information about this radio's features"""
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank = False
        rf.has_tuning_step = False
        rf.has_name = False
        rf.has_offset = True
        rf.has_mode = True
        rf.has_dtcs = True
        rf.has_rx_dtcs = True
        rf.has_dtcs_polarity = False
        rf.has_ctone = True
        rf.has_cross = True
        rf.valid_duplexes = ["", "-", "+", "off"]
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_cross_modes = [
            "Tone->Tone",
            "DTCS->DTCS",
            "DTCS->",
            "->DTCS"]
        rf.valid_dtcs_codes = DTCS_CODES
        rf.valid_skips = SKIP_VALUES
        rf.valid_modes = ["FM"]
        rf.valid_power_levels = POWER_LEVELS
        # rf.valid_tuning_steps = [5.0]
        rf.valid_bands = [self._range]
        rf.memory_bounds = (1, self._upper)
        return rf

    def sync_in(self):
        """Do a download of the radio eeprom"""
        try:
            data = _do_download(self)
        except Exception as e:
            raise errors.RadioError("Failed to communicate with radio:\n %s" % e)

        # match model
        if _model_match(self, data) is False:
            raise errors.RadioError("Incorrect radio model")

        self._mmap = memmap.MemoryMap(data)
        self.process_mmap()

        # set the channel count from the radio eeprom
        self._upper = int(ord(data[4]))

    def sync_out(self):
        """Do an upload to the radio eeprom"""
        # update checksum
        _update_cs(self)

        # sanity check, match model
        data = str(self.get_mmap())
        if len(data) != MEM_SIZE:
            raise errors.RadioError("Wrong radio image? Size miss match.")

        if _model_match(self, data) is False:
            raise errors.RadioError("Wrong image? Fingerprint miss match")

        try:
            _do_upload(self)
        except Exception as e:
            msg = "Failed to communicate with radio:\n%s" % e
            raise errors.RadioError(msg)

    def process_mmap(self):
        """Process the memory object"""
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_raw_memory(self, number):
        """Return a raw representation of the memory object"""
        return repr(self._memobj.memory[number])

    def _decode_tone(self, mem, rx=True):
        """Parse the tone data to decode from mem tones are encoded like this
        CTCS: mapped [0x80...0xa5] = [67.0...250.3]
        DTCS: mixed  [0x88, 0x23] last is BCD and first is the 100 power - 88

        It return: ((''|DTCS|Tone), Value (None|###), None)"""
        mode = ""
        tone = None

        # get the tone depending of rx or tx
        if rx:
            t = mem.rx_tone
        else:
            t = mem.tx_tone

        tMSB = t[0]
        tLSB = t[1]

        # no tone at all
        if (tMSB == 0 and tLSB < 128):
            return ('', None, None)

        # extract the tone info
        if tMSB == 0x00:
            # CTCS
            mode = "Tone"
            try:
                tone = TONES[tLSB - 128]
            except IndexError:
                LOG.debug("Error decoding a CTCS tone")
                pass
        else:
            # DTCS
            mode = "DTCS"
            try:
                tone = ((tMSB - 0x88) * 100) + bcd_to_int(tLSB)
            except IndexError:
                LOG.debug("Error decoding a DTCS tone")
                pass

        return (mode, tone, None)

    def _encode_tone(self, mem, mode, value, pol, rx=True):
        """Parse the tone data to encode from UI to mem
        CTCS: mapped [0x80...0xa5] = [67.0...250.3]
        DTCS: mixed  [0x88, 0x23] last is BCD and first is the 100 power - 88
        """

        # array to pass
        tone = [0x00, 0x00]

        # which mod
        if mode == "DTCS":
            tone[0] = int(value / 100) + 0x88
            tone[1] = int_to_bcd(value % 100)

        if mode == "Tone":
            # CTCS
            tone[1] = TONES.index(value) + 128

        # set it
        if rx:
            mem.rx_tone = tone
        else:
            mem.tx_tone = tone

    def get_memory(self, number):
        """Extract a memory object from the memory map"""
        # Get a low-level memory object mapped to the image
        _mem = self._memobj.memory[number - 1]
        # Create a high-level memory object to return to the UI
        mem = chirp_common.Memory()
        # number
        mem.number = number

        # empty
        if bool(_mem.empty) is True:
            mem.empty = True
            return mem

        # rx freq
        mem.freq = int(_mem.rx_freq) * 1000

        # power
        mem.power = POWER_LEVELS[int(_mem.power)]

        # checking if tx freq is disabled
        if bool(_mem.notx) is True:
            mem.duplex = "off"
            mem.offset = 0
        else:
            tx = int(_mem.tx_freq) * 1000
            if tx == mem.freq:
                mem.offset = 0
                mem.duplex = ""
            else:
                mem.duplex = mem.freq > tx and "-" or "+"
                mem.offset = abs(tx - mem.freq)

        # skip
        mem.skip = SKIP_VALUES[_mem.skip]

        # tone data
        rxtone = txtone = None
        rxtone = self._decode_tone(_mem)
        txtone = self._decode_tone(_mem, False)
        chirp_common.split_tone_decode(mem, txtone, rxtone)

        # this radio has a primitive mode to show the channel number on a 7-segment
        # two digit LCD, we will use channel number
        # we will use a trick to show the numbers < 10 with a space not a zero in front
        chname = int_to_bcd(mem.number)
        if mem.number < 10:
            # convert to F# as BCD
            chname = mem.number + 240

        _mem.chname = chname

        # Extra
        mem.extra = RadioSettingGroup("extra", "Extra")

        # bcl preparations: ["OFF", "Carrier", "Tone"]
        bcls = 0
        if _mem.bclo_cw:
            bcls = 1
        if _mem.bclo_tone:
            bcls = 2

        bcl = RadioSetting("bclo", "Busy channel lockout",
                           RadioSettingValueList(LIST_BCL,
                                                 LIST_BCL[bcls]))
        mem.extra.append(bcl)

        # return mem
        return mem

    def set_memory(self, mem):
        """Store details about a high-level memory to the memory map
        This is called when a user edits a memory in the UI"""
        # Get a low-level memory object mapped to the image
        _mem = self._memobj.memory[mem.number - 1]

        # Empty memory
        _mem.empty = mem.empty
        if mem.empty:
            _mem.rx_freq = _mem.tx_freq = 0
            return

        # freq rx
        _mem.rx_freq = mem.freq / 1000

        # power, # default power level is high
        _mem.power = 0 if mem.power is None else POWER_LEVELS.index(mem.power)

        # freq tx
        if mem.duplex == "+":
            _mem.tx_freq = (mem.freq + mem.offset) / 1000
        elif mem.duplex == "-":
            _mem.tx_freq = (mem.freq - mem.offset) / 1000
        elif mem.duplex == "off":
            _mem.notx = 1
            _mem.tx_freq = _mem.rx_freq
        else:
            _mem.tx_freq = mem.freq / 1000

        # scan add property
        _mem.skip = SKIP_VALUES.index(mem.skip)

        # tone data
        ((txmode, txtone, txpol), (rxmode, rxtone, rxpol)) = \
            chirp_common.split_tone_encode(mem)

        # validate tone data from here
        if rxmode == "Tone" and rxtone in invalid_tones:
            msg = "The tone %s Hz is not valid for this radio" % rxtone
            raise errors.UnsupportedToneError(msg)

        if txmode == "Tone" and txtone in invalid_tones:
            msg = "The tone %s Hz is not valid for this radio" % txtone
            raise errors.UnsupportedToneError(msg)

        if rxmode == "DTCS" and rxtone not in DTCS_CODES:
            msg = "The digital tone %s is not valid for this radio" % rxtone
            raise errors.UnsupportedToneError(msg)

        if txmode == "DTCS" and txtone not in DTCS_CODES:
            msg = "The digital tone %s is not valid for this radio" % txtone
            raise errors.UnsupportedToneError(msg)

        self._encode_tone(_mem, rxmode, rxtone, rxpol)
        self._encode_tone(_mem, txmode, txtone, txpol, False)

        # this radio has a primitive mode to show the channel number on a 7-segment
        # two digit LCD, we will use channel number
        # we will use a trick to show the numbers < 10 with a space not a zero in front
        chname = int_to_bcd(mem.number)
        if mem.number < 10:
            # convert to F# as BCD
            chname = mem.number + 240

        def _zero_settings():
            _mem.bclo_cw = 0
            _mem.bclo_tone = 0

        # extra settings
        if len(mem.extra) > 0:
            # there are setting, parse
            LOG.debug("Extra-Setting supplied. Setting them.")
            # Zero them all first so any not provided by model don't
            # stay set
            _zero_settings()
            for setting in mem.extra:
                if setting.get_name() == "bclo":
                    sw = LIST_BCL.index(str(setting.value))
                    if sw == 0:
                        # empty
                        _zero_settings()
                    if sw == 1:
                        # carrier
                        _mem.bclo_cw = 1
                    if sw == 2:
                        # tone
                        _mem.bclo_tone = 1
                        # activate the tone
                        _mem.rx_tone = [0x00, 0x80]
        else:
            # reset extra settings
            _zero_settings()

        _mem.chname = chname

        return mem

    def get_settings(self):
        _settings = self._memobj.settings
        basic = RadioSettingGroup("basic", "Basic Settings")
        group = RadioSettings(basic)

        # ## Basic Settings
        scanr = RadioSetting("scan_resume", "Scan resume by",
                             RadioSettingValueList(
                              LIST_SCAN_RESUME, LIST_SCAN_RESUME[_settings.scan_resume]))
        basic.append(scanr)

        scant = RadioSetting("scan_time", "Scan time per channel",
                             RadioSettingValueList(
                              LIST_SCAN_TIME, LIST_SCAN_TIME[_settings.scan_time]))
        basic.append(scant)

        LIST_PCH = ["%s" % x for x in range(1, _settings.chcount + 1)]
        pch1 = RadioSetting("pch1", "Priority channel 1",
                            RadioSettingValueList(
                                 LIST_PCH, LIST_PCH[_settings.pch1]))
        basic.append(pch1)

        pch2 = RadioSetting("pch2", "Priority channel 2",
                            RadioSettingValueList(
                                LIST_PCH, LIST_PCH[_settings.pch2]))
        basic.append(pch2)

        scanp = RadioSetting("priority_during_scan", "Disable priority during scan",
                             RadioSettingValueBoolean(_settings.priority_during_scan))
        basic.append(scanp)

        scanps = RadioSetting("priority_speed", "Priority scan speed",
                              RadioSettingValueList(
                                LIST_SCAN_P_SPEED, LIST_SCAN_P_SPEED[_settings.priority_speed]))
        basic.append(scanps)

        oh = RadioSetting("off_hook", "Off Hook",  # inverted
                          RadioSettingValueBoolean(not _settings.off_hook))
        basic.append(oh)

        tb = RadioSetting("talk_back", "Talk Back",  # inverted
                          RadioSettingValueBoolean(not _settings.talk_back))
        basic.append(tb)

        tot = RadioSetting("tot", "Time out timer",
                           RadioSettingValueList(
                                 LIST_TOT, LIST_TOT[_settings.tot]))
        basic.append(tot)

        totr = RadioSetting("tot_resume", "Time out timer resume guard",
                            RadioSettingValueList(
                               LIST_TOT_RESUME, LIST_TOT_RESUME[_settings.tot_resume]))
        basic.append(totr)

        ak = RadioSetting("a_key", "A Key function",
                          RadioSettingValueList(
                                LIST_A_KEY, LIST_A_KEY[_settings.a_key]))
        basic.append(ak)

        monitor = RadioSetting("monitor", "Monitor",  # inverted
                               RadioSettingValueBoolean(not _settings.monitor))
        basic.append(monitor)

        homec = RadioSetting("home_channel", "Home Channel is",
                             RadioSettingValueList(
                                 LIST_HOME_CHANNEL, LIST_HOME_CHANNEL[_settings.home_channel]))
        basic.append(homec)

        txd = RadioSetting("tx_carrier_delay", "Talk Back",
                           RadioSettingValueBoolean(_settings.tx_carrier_delay))
        basic.append(txd)

        return group

    def set_settings(self, uisettings):
        _settings = self._memobj.settings

        for element in uisettings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            if not element.changed():
                continue

            try:
                name = element.get_name()
                value = element.value

                obj = getattr(_settings, name)
                if name in ["off_hook", "talk_back", "monitor"]:
                    setattr(_settings, name, not value)
                else:
                    setattr(_settings, name, value)

                LOG.debug("Setting %s: %s" % (name, value))
            except Exception:
                LOG.debug(element.get_name())
                raise

    @classmethod
    def match_model(cls, filedata, filename):
        match_size = False
        match_model = False

        # testing the file data size
        if len(filedata) == cls._memsize:
            match_size = True

        # testing the firmware fingerprint, this experimental
        try:
            match_model = _model_match(cls, filedata)
        except Exception:
            match_model = False

        return match_size and match_model


@directory.register
class ftl1011(ftlx011):
    """Vertex FTL-1011"""
    MODEL = "FTL-1011"
    _memsize = MEM_SIZE
    _upper = 4
    _range = [44000000, 56000000]
    finger = ["\x52", "\x16\x90"]


@directory.register
class ftl2011(ftlx011):
    """Vertex FTL-2011"""
    MODEL = "FTL-2011"
    _memsize = MEM_SIZE
    _upper = 24
    _range = [134000000, 174000000]
    finger = ["\x50", "\x21\x40"]


@directory.register
class ftl7011(ftlx011):
    """Vertex FTL-7011"""
    MODEL = "FTL-7011"
    _memsize = MEM_SIZE
    _upper = 24
    _range = [400000000, 512000000]
    finger = ["\x54", "\x47\x90"]


@directory.register
class ftl8011(ftlx011):
    """Vertex FTL-8011"""
    MODEL = "FTL-8011"
    _memsize = MEM_SIZE
    _upper = 24
    _range = [400000000, 512000000]
    finger = ["\x5c", "\x45\x10"]



# Copyright 2011 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
import logging

from chirp.drivers import yaesu_clone
from chirp import chirp_common, memmap, bitwise, directory, errors
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueList, RadioSettingValueString, \
    RadioSettings

LOG = logging.getLogger(__name__)

ACK = b"\x06"


def _send(pipe, data):
    pipe.write(data)
    echo = pipe.read(len(data))
    if echo != data:
        raise errors.RadioError("Error reading echo (Bad cable?)")


def _download(radio):
    data = b""
    for i in range(0, 10):
        chunk = radio.pipe.read(8)
        if len(chunk) == 8:
            data += chunk
            break
        elif chunk:
            raise Exception("Received invalid response from radio")
        time.sleep(1)
        LOG.info("Trying again...")

    if not data:
        raise Exception("Radio is not responding")

    _send(radio.pipe, ACK)

    for i in range(0, 448):
        chunk = radio.pipe.read(64)
        data += chunk
        _send(radio.pipe, ACK)
        if len(chunk) == 1 and i == 447:
            break
        elif len(chunk) != 64:
            raise Exception("Reading block %i was short (%i)" %
                            (i, len(chunk)))
        if radio.status_fn:
            status = chirp_common.Status()
            status.cur = i * 64
            status.max = radio.get_memsize()
            status.msg = "Cloning from radio"
            radio.status_fn(status)

    return memmap.MemoryMapBytes(data)


def _upload(radio):
    _send(radio.pipe, radio.get_mmap()[0:8])

    ack = radio.pipe.read(1)
    if ack != ACK:
        raise Exception("Radio did not respond")

    for i in range(0, 448):
        offset = 8 + (i * 64)
        _send(radio.pipe, radio.get_mmap()[offset:offset + 64])
        ack = radio.pipe.read(1)
        if ack != ACK:
            raise Exception(_("Radio did not ack block %i") % i)

        if radio.status_fn:
            status = chirp_common.Status()
            status.cur = offset + 64
            status.max = radio.get_memsize()
            status.msg = "Cloning to radio"
            radio.status_fn(status)


def _decode_freq(freqraw):
    freq = int(freqraw) * 10000
    if freq > 8000000000:
        freq = (freq - 8000000000) + 5000

    if freq > 4000000000:
        freq -= 4000000000
        for i in range(0, 3):
            freq += 2500
            if chirp_common.required_step(freq) == 12.5:
                break

    return freq


def _encode_freq(freq):
    freqraw = freq / 10000
    flags = 0x00
    if ((freq / 1000) % 10) >= 5:
        flags += 0x80
    if chirp_common.is_fractional_step(freq):
        flags += 0x40
    return freqraw, flags


def _decode_name(mem):
    name = ""
    for i in mem:
        if i == 0xFF:
            break
        try:
            name += CHARSET[i]
        except IndexError:
            LOG.error("Unknown char index: %i " % (i))
    return name


def _encode_name(mem):
    name = [None] * 6
    for i in range(0, 6):
        try:
            name[i] = CHARSET.index(mem[i])
        except IndexError:
            name[i] = CHARSET.index(" ")

    return name


MEM_FORMAT = """
#seekto 0x0024;
struct {
  u8 apo;
  u8 x25:3,
     tot:5;
  u8 x26;
  u8 x27;
  u8 x28:4,
     rf_sql:4;
  u8 x29:4,
     int_cd:4;
  u8 x2A:4,
     int_mr:4;
  u8 x2B:5,
     lock:3;
  u8 x2C:5,
     dt_dly:3;
  u8 x2D:7,
     dt_spd:1;
  u8 ar_bep;
  u8 x2F:6,
     lamp:2;
  u8 x30:5,
     bell:3;
  u8 x31:5,
     rxsave:3;
  u8 x32;
  u8 x33;
  u8 x34;
  u8 x35;
  u8 x36;
  u8 x37;
  u8 wx_alt:1,
     x38_1:3,
     ar_int:1,
     x38_5:3;
  u8 x39:3,
     ars:1,
     vfo_bnd:1,
     dcs_nr:2,
     ssrch:1;
  u8 pri_rvt:1,
     x3A_1:1,
     beep_sc:1,
     edg_bep:1,
     beep_key:1,
     inet:2,
     x3A_7:1;
  u8 x3B_0:5,
     scn_md:1,
     x3B_6:2;
  u8 x3C_0:2,
     rev_hm:1,
     mt_cl:1,
     resume:2,
     txsave:1,
     pag_abk:1;
  u8 x3D_0:1,
     scn_lmp:1,
     x3D_2:1,
     bsy_led:1,
     x3D_4:1,
     tx_led:1,
     x3D_6:2;
  u8 x3E_0:2,
     bclo:1,
     x3E_3:5;
} settings;

#seekto 0x09E;
ul16 mbs;

#seekto 0x0C8;
struct {
  u8 memory[16];
} dtmf[9];

struct mem {
  u8 used:1,
     unknown1:1,
     isnarrow:1,
     isam:1,
     duplex:4;
  bbcd freq[3];
  u8 unknown2:1,
     step:3,
     unknown2_1:1,
     tmode:3;
  bbcd tx_freq[3];
  u8 power:2,
     tone:6;
  u8 unknown4:1,
     dtcs:7;
  u8 unknown5;
  u16 unknown5_1:1,
      offset:15;
  u8 unknown6[3];
};

#seekto 0x0248;
struct mem memory[1000];

#seekto 0x40c8;
struct mem pms[100];

#seekto 0x6EC8;
// skips:2 for Memory M in [1, 1000] is in flags[(M-1)/4].skip((M-1)%4).
// Interpret with SKIPS[].
// PMS memories L0 - U50 aka memory 1001 - 1100 don't have skip flags.
struct {
  u8 skip3:2,
     skip2:2,
     skip1:2,
     skip0:2;
} flags[250];

#seekto 0x4708;
struct {
  u8 name[6];
  u8 use_name:1,
     unknown1:7;
  u8 valid:1,
     unknown2:7;
} names[1000];

#seekto 0x69C8;
struct {
  bbcd memory[128];
} banks[10];

#seekto 0x6FC8;
u8 checksum;
"""

DUPLEX = ["", "", "-", "+", "split", "off"]
TMODES = ["", "Tone", "TSQL", "TSQL-R", "DTCS",
          "DTCS->", "Tone->DTCS", "DTCS->Tone"]
POWER_LEVELS = [chirp_common.PowerLevel("High", watts=5.0),
                chirp_common.PowerLevel("Mid", watts=2.0),
                chirp_common.PowerLevel("Low", watts=0.5)]
STEPS = [5.0, 10.0, 12.5, 15.0, 20.0, 25.0, 50.0, 100.0]
SKIPS = ["", "S", "P"]
DTMF_CHARS = list("0123456789ABCD*#")
CHARSET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ !`o$%&'()*+,-./|;/=>?@[u]^_"
SPECIALS = ["%s%d" % (c, i + 1) for i in range(0, 50) for c in ('L', 'U')]


class FT60BankModel(chirp_common.BankModel):

    def get_num_mappings(self):
        return 10

    def get_mappings(self):
        banks = []
        for i in range(0, self.get_num_mappings()):
            bank = chirp_common.Bank(self, "%i" % (i + 1), "Bank %i" % (i + 1))
            bank.index = i
            banks.append(bank)
        return banks

    def add_memory_to_mapping(self, memory, bank):
        number = (memory.number - 1) / 8
        mask = 1 << ((memory.number - 1) & 7)
        self._radio._memobj.banks[bank.index].memory[number].set_bits(mask)

    def remove_memory_from_mapping(self, memory, bank):
        number = (memory.number - 1) / 8
        mask = 1 << ((memory.number - 1) & 7)
        m = self._radio._memobj.banks[bank.index].memory[number]
        if m.get_bits(mask) != mask:
            raise Exception("Memory %i is not in bank %s." %
                            (memory.number, bank))
        self._radio._memobj.banks[bank.index].memory[number].clr_bits(mask)

    def get_mapping_memories(self, bank):
        memories = []
        for i in range(*self._radio.get_features().memory_bounds):
            number = (i - 1) / 8
            mask = 1 << ((i - 1) & 7)
            m = self._radio._memobj.banks[bank.index].memory[number]
            if m.get_bits(mask) == mask:
                memories.append(self._radio.get_memory(i))
        return memories

    def get_memory_mappings(self, memory):
        banks = []
        for bank in self.get_mappings():
            number = (memory.number - 1) / 8
            mask = 1 << ((memory.number - 1) & 7)
            m = self._radio._memobj.banks[bank.index].memory[number]
            if m.get_bits(mask) == mask:
                banks.append(bank)
        return banks


@directory.register
class FT60Radio(yaesu_clone.YaesuCloneModeRadio):

    """Yaesu FT-60"""
    BAUD_RATE = 9600
    VENDOR = "Yaesu"
    MODEL = "FT-60"
    _model = b"AH017"

    _memsize = 28617

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.pre_download = _(
            "1. Turn radio off.\n"
            "2. Connect cable to MIC/SP jack.\n"
            "3. Press and hold in the [MONI] switch while turning the\n"
            "     radio on.\n"
            "4. Rotate the DIAL job to select \"F8 CLONE\".\n"
            "5. Press the [F/W] key momentarily.\n"
            "6. <b>After clicking OK</b>, hold the [PTT] switch\n"
            "     for one second to send image.\n")
        rp.pre_upload = _(
            "1. Turn radio off.\n"
            "2. Connect cable to MIC/SP jack.\n"
            "3. Press and hold in the [MONI] switch while turning the\n"
            "     radio on.\n"
            "4. Rotate the DIAL job to select \"F8 CLONE\".\n"
            "5. Press the [F/W] key momentarily.\n"
            "6. Press the [MONI] switch (\"--RX--\" will appear on the"
            " LCD).\n")
        return rp

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.memory_bounds = (1, 1000)
        rf.valid_duplexes = DUPLEX[1:]
        rf.valid_tmodes = [x for x in TMODES if '->' not in x] + ['Cross']
        rf.valid_cross_modes = ['Tone->DTCS', 'DTCS->Tone', 'DTCS->']
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_tuning_steps = STEPS
        rf.valid_skips = SKIPS
        rf.valid_special_chans = SPECIALS
        rf.valid_characters = CHARSET
        rf.valid_name_length = 6
        rf.valid_modes = ["FM", "NFM", "AM"]
        rf.valid_bands = [(108000000, 520000000), (700000000, 999990000)]
        rf.can_odd_split = True
        rf.has_ctone = False
        rf.has_bank = True
        rf.has_settings = True
        rf.has_dtcs_polarity = False

        return rf

    def get_bank_model(self):
        return FT60BankModel(self)

    def _checksums(self):
        return [yaesu_clone.YaesuChecksum(0x0000, 0x6FC7)]

    def sync_in(self):
        try:
            self._mmap = _download(self)
        except errors.RadioError:
            raise
        except Exception as e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)
        self.process_mmap()
        self.check_checksums()

    def sync_out(self):
        self.update_checksums()
        try:
            _upload(self)
        except errors.RadioError:
            raise
        except Exception as e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_settings(self):
        _settings = self._memobj.settings

        repeater = RadioSettingGroup("repeater", "Repeater Settings")
        ctcss = RadioSettingGroup("ctcss", "CTCSS/DCS/DTMF Settings")
        arts = RadioSettingGroup("arts", "ARTS Settings")
        scan = RadioSettingGroup("scan", "Scan Settings")
        power = RadioSettingGroup("power", "Power Saver Settings")
        wires = RadioSettingGroup("wires", "WiRES(tm) Settings")
        eai = RadioSettingGroup("eai", "EAI/EPCS Settings")
        switch = RadioSettingGroup("switch", "Switch/Knob Settings")
        misc = RadioSettingGroup("misc", "Miscellaneous Settings")
        mbls = RadioSettingGroup("banks", "Memory Bank Link Scan")

        setmode = RadioSettings(repeater, ctcss, arts, scan, power,
                                wires, eai, switch, misc, mbls)

        # APO
        opts = ["OFF"] + ["%0.1f" % (x * 0.5) for x in range(1, 24 + 1)]
        misc.append(
            RadioSetting(
                "apo", "Automatic Power Off",
                RadioSettingValueList(opts, opts[_settings.apo])))

        # AR.BEP
        opts = ["OFF", "INRANG", "ALWAYS"]
        arts.append(
            RadioSetting(
                "ar_bep", "ARTS Beep",
                RadioSettingValueList(opts, opts[_settings.ar_bep])))

        # AR.INT
        opts = ["25 SEC", "15 SEC"]
        arts.append(
            RadioSetting(
                "ar_int", "ARTS Polling Interval",
                RadioSettingValueList(opts, opts[_settings.ar_int])))

        # ARS
        opts = ["OFF", "ON"]
        repeater.append(
            RadioSetting(
                "ars", "Automatic Repeater Shift",
                RadioSettingValueList(opts, opts[_settings.ars])))

        # BCLO
        opts = ["OFF", "ON"]
        misc.append(RadioSetting(
            "bclo", "Busy Channel Lock-Out",
            RadioSettingValueList(opts, opts[_settings.bclo])))

        # BEEP
        opts = ["OFF", "KEY", "KEY+SC"]
        rs = RadioSetting(
            "beep_key", "Enable the Beeper",
            RadioSettingValueList(
                opts, opts[_settings.beep_key + _settings.beep_sc]))

        def apply_beep(s, obj):
            setattr(obj, "beep_key",
                    (int(s.value) & 1) or ((int(s.value) >> 1) & 1))
            setattr(obj, "beep_sc", (int(s.value) >> 1) & 1)
        rs.set_apply_callback(apply_beep, self._memobj.settings)
        switch.append(rs)

        # BELL
        opts = ["OFF", "1T", "3T", "5T", "8T", "CONT"]
        ctcss.append(RadioSetting("bell", "Bell Repetitions",
                                  RadioSettingValueList(opts, opts[
                                                        _settings.bell])))

        # BSY.LED
        opts = ["ON", "OFF"]
        misc.append(RadioSetting("bsy_led", "Busy LED",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.bsy_led])))

        # DCS.NR
        opts = ["TR/X N", "RX R", "TX R", "T/RX R"]
        ctcss.append(RadioSetting("dcs_nr", "\"Inverted\" DCS Code Decoding",
                                  RadioSettingValueList(opts, opts[
                                                        _settings.dcs_nr])))

        # DT.DLY
        opts = ["50 ms", "100 ms", "250 ms", "450 ms", "750 ms", "1000 ms"]
        ctcss.append(RadioSetting("dt_dly", "DTMF Autodialer Delay Time",
                                  RadioSettingValueList(opts, opts[
                                                        _settings.dt_dly])))

        # DT.SPD
        opts = ["50 ms", "100 ms"]
        ctcss.append(RadioSetting("dt_spd", "DTMF Autodialer Sending Speed",
                                  RadioSettingValueList(opts, opts[
                                                        _settings.dt_spd])))

        # DT.WRT
        for i in range(0, 9):
            dtmf = self._memobj.dtmf[i]
            str = ""
            for c in dtmf.memory:
                if c == 0xFF:
                    break
                if c < len(DTMF_CHARS):
                    str += DTMF_CHARS[c]
            val = RadioSettingValueString(0, 16, str, False)
            val.set_charset(DTMF_CHARS + list("abcd"))
            rs = RadioSetting("dtmf_%i" % i,
                              "DTMF Autodialer Memory %i" % (i + 1), val)

            def apply_dtmf(s, obj):
                str = s.value.get_value().upper().rstrip()
                val = [DTMF_CHARS.index(x) for x in str]
                for x in range(len(val), 16):
                    val.append(0xFF)
                obj.memory = val
            rs.set_apply_callback(apply_dtmf, dtmf)
            ctcss.append(rs)

        # EDG.BEP
        opts = ["OFF", "ON"]
        misc.append(RadioSetting("edg_bep", "Band Edge Beeper",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.edg_bep])))

        # I.NET
        opts = ["OFF", "COD", "MEM"]
        rs = RadioSetting("inet", "Internet Link Connection",
                          RadioSettingValueList(
                              opts, opts[_settings.inet - 1]))

        def apply_inet(s, obj):
            setattr(obj, s.get_name(), int(s.value) + 1)
        rs.set_apply_callback(apply_inet, self._memobj.settings)
        wires.append(rs)

        # INT.CD
        opts = ["CODE 0", "CODE 1", "CODE 2", "CODE 3", "CODE 4",
                "CODE 5", "CODE 6", "CODE 7", "CODE 8", "CODE 9",
                "CODE A", "CODE B", "CODE C", "CODE D", "CODE E", "CODE F"]
        wires.append(RadioSetting("int_cd", "Access Number for WiRES(TM)",
                                  RadioSettingValueList(opts, opts[
                                                        _settings.int_cd])))

        # INT.MR
        opts = ["d1", "d2", "d3", "d4", "d5", "d6", "d7", "d8", "d9"]
        wires.append(RadioSetting(
            "int_mr", "Access Number (DTMF) for Non-WiRES(TM)",
                     RadioSettingValueList(opts, opts[_settings.int_mr])))

        # LAMP
        opts = ["KEY", "5SEC", "TOGGLE"]
        switch.append(RadioSetting("lamp", "Lamp Mode",
                                   RadioSettingValueList(opts, opts[
                                                         _settings.lamp])))

        # LOCK
        opts = ["LK KEY", "LKDIAL", "LK K+D", "LK PTT",
                "LK P+K", "LK P+D", "LK ALL"]
        rs = RadioSetting("lock", "Control Locking",
                          RadioSettingValueList(
                              opts, opts[_settings.lock - 1]))

        def apply_lock(s, obj):
            setattr(obj, s.get_name(), int(s.value) + 1)
        rs.set_apply_callback(apply_lock, self._memobj.settings)
        switch.append(rs)

        # M/T-CL
        opts = ["MONI", "T-CALL"]
        switch.append(RadioSetting("mt_cl", "MONI Switch Function",
                                   RadioSettingValueList(opts, opts[
                                                         _settings.mt_cl])))

        # PAG.ABK
        opts = ["OFF", "ON"]
        eai.append(RadioSetting("pag_abk", "Paging Answer Back",
                                RadioSettingValueList(opts, opts[
                                                      _settings.pag_abk])))

        # RESUME
        opts = ["TIME", "HOLD", "BUSY"]
        scan.append(RadioSetting("resume", "Scan Resume Mode",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.resume])))

        # REV/HM
        opts = ["REV", "HOME"]
        switch.append(RadioSetting("rev_hm", "HM/RV Key Function",
                                   RadioSettingValueList(opts, opts[
                                                         _settings.rev_hm])))

        # RF.SQL
        opts = ["OFF", "S-1", "S-2", "S-3", "S-4", "S-5", "S-6",
                "S-7", "S-8", "S-FULL"]
        misc.append(RadioSetting("rf_sql", "RF Squelch Threshold",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.rf_sql])))

        # PRI.RVT
        opts = ["OFF", "ON"]
        scan.append(RadioSetting("pri_rvt", "Priority Revert",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.pri_rvt])))

        # RXSAVE
        opts = ["OFF", "200 ms", "300 ms", "500 ms", "1 s", "2 s"]
        power.append(RadioSetting(
            "rxsave", "Receive Mode Battery Savery Interval",
                     RadioSettingValueList(opts, opts[_settings.rxsave])))

        # S.SRCH
        opts = ["SINGLE", "CONT"]
        misc.append(RadioSetting("ssrch", "Smart Search Sweep Mode",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.ssrch])))

        # SCN.MD
        opts = ["MEM", "ONLY"]
        scan.append(RadioSetting(
                    "scn_md", "Memory Scan Channel Selection Mode",
                    RadioSettingValueList(opts, opts[_settings.scn_md])))

        # SCN.LMP
        opts = ["OFF", "ON"]
        scan.append(RadioSetting("scn_lmp", "Scan Lamp",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.scn_lmp])))

        # TOT
        opts = ["OFF"] + ["%dMIN" % (x) for x in range(1, 30 + 1)]
        misc.append(RadioSetting("tot", "Timeout Timer",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.tot])))

        # TX.LED
        opts = ["ON", "OFF"]
        misc.append(RadioSetting("tx_led", "TX LED",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.tx_led])))

        # TXSAVE
        opts = ["OFF", "ON"]
        power.append(RadioSetting("txsave", "Transmitter Battery Saver",
                                  RadioSettingValueList(opts, opts[
                                                        _settings.txsave])))

        # VFO.BND
        opts = ["BAND", "ALL"]
        misc.append(RadioSetting("vfo_bnd", "VFO Band Edge Limiting",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.vfo_bnd])))

        # WX.ALT
        opts = ["OFF", "ON"]
        scan.append(RadioSetting("wx_alt", "Weather Alert Scan",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.wx_alt])))

        # MBS
        for i in range(0, 10):
            opts = ["OFF", "ON"]
            mbs = (self._memobj.mbs >> i) & 1
            rs = RadioSetting("mbs%i" % i, "Bank %s Scan" % (i + 1),
                              RadioSettingValueList(opts, opts[mbs]))

            def apply_mbs(s, index):
                if int(s.value):
                    self._memobj.mbs |= (1 << index)
                else:
                    self._memobj.mbs &= ~(1 << index)
            rs.set_apply_callback(apply_mbs, i)
            mbls.append(rs)

        return setmode

    def set_settings(self, uisettings):
        _settings = self._memobj.settings
        for element in uisettings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            if not element.changed():
                continue

            try:
                name = element.get_name()
                value = element.value

                if element.has_apply_callback():
                    LOG.debug("Using apply callback")
                    element.run_apply_callback()
                else:
                    obj = getattr(_settings, name)
                    setattr(_settings, name, value)

                LOG.debug("Setting %s: %s" % (name, value))
            except Exception:
                LOG.debug(element.get_name())
                raise

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number - 1]) + \
            repr(self._memobj.flags[(number - 1) / 4]) + \
            repr(self._memobj.names[number - 1])

    def get_memory(self, number):

        mem = chirp_common.Memory()

        if isinstance(number, str):
            # pms channel
            mem.number = 1001 + SPECIALS.index(number)
            mem.extd_number = number
            mem.immutable = ["number", "extd_number", "name", "skip"]
            _mem = self._memobj.pms[mem.number - 1001]
            _nam = _skp = None
        elif number > 1000:
            # pms channel
            mem.number = number
            mem.extd_number = SPECIALS[number - 1001]
            mem.immutable = ["number", "extd_number", "name", "skip"]
            _mem = self._memobj.pms[mem.number - 1001]
            _nam = _skp = None
        else:
            mem.number = number
            _mem = self._memobj.memory[mem.number - 1]
            _nam = self._memobj.names[mem.number - 1]
            _skp = self._memobj.flags[(mem.number - 1) / 4]

        if not _mem.used:
            mem.empty = True
            return mem

        mem.freq = _decode_freq(_mem.freq)
        mem.offset = int(_mem.offset) * 50000
        mem.duplex = DUPLEX[_mem.duplex]
        if mem.duplex == "split":
            if int(_mem.tx_freq) == 0:
                mem.duplex = "off"
            else:
                mem.offset = _decode_freq(_mem.tx_freq)
        tmode = TMODES[_mem.tmode]
        if '->' in tmode:
            mem.tmode = 'Cross'
            mem.cross_mode = tmode
        else:
            mem.tmode = tmode
        mem.rtone = chirp_common.TONES[_mem.tone]
        mem.dtcs = chirp_common.DTCS_CODES[_mem.dtcs]
        mem.power = POWER_LEVELS[_mem.power]
        mem.mode = _mem.isam and "AM" or _mem.isnarrow and "NFM" or "FM"
        mem.tuning_step = STEPS[_mem.step]

        if _skp is not None:
            skip = _skp["skip%i" % ((mem.number - 1) % 4)]
            mem.skip = SKIPS[skip]

        if _nam is not None:
            if _nam.use_name and _nam.valid:
                mem.name = _decode_name(_nam.name).rstrip()

        return mem

    def set_memory(self, mem):

        if mem.number > 1000:
            # pms channel
            _mem = self._memobj.pms[mem.number - 1001]
            _nam = _skp = None
        else:
            _mem = self._memobj.memory[mem.number - 1]
            _nam = self._memobj.names[mem.number - 1]
            _skp = self._memobj.flags[(mem.number - 1) / 4]

        assert (_mem)
        if mem.empty:
            _mem.used = False
            return

        if not _mem.used:
            _mem.set_raw("\x00" * 16)
            _mem.used = 1

        _mem.freq, flags = _encode_freq(mem.freq)
        _mem.freq[0].set_bits(flags)
        if mem.duplex == "split":
            _mem.tx_freq, flags = _encode_freq(mem.offset)
            _mem.tx_freq[0].set_bits(flags)
            _mem.offset = 0
        elif mem.duplex == "off":
            _mem.tx_freq = 0
            _mem.offset = 0
        else:
            _mem.tx_freq = 0
            _mem.offset = mem.offset / 50000
        _mem.duplex = DUPLEX.index(mem.duplex)
        if mem.tmode == 'Cross':
            _mem.tmode = TMODES.index(mem.cross_mode)
        else:
            _mem.tmode = TMODES.index(mem.tmode)
        _mem.tone = chirp_common.TONES.index(mem.rtone)
        _mem.dtcs = chirp_common.DTCS_CODES.index(mem.dtcs)
        _mem.power = mem.power and POWER_LEVELS.index(mem.power) or 0
        _mem.isnarrow = mem.mode == "NFM"
        _mem.isam = mem.mode == "AM"
        _mem.step = STEPS.index(mem.tuning_step)

        if _skp is not None:
            _skp["skip%i" % ((mem.number - 1) % 4)] = SKIPS.index(mem.skip)

        if _nam is not None:
            _nam.name = _encode_name(mem.name)
            _nam.use_name = mem.name.strip() and True or False
            _nam.valid = _nam.use_name
